"""Common drawing functions"""
import math


class Pixel(object):
    """Pixel class"""
    def __init__(self, driver):
        self.driver = driver
        self.options['color'] = {
            'R': 255, 'G': 255, 'B': 255
        }
        self.options['background_color'] = {
            'R': 0, 'G': 0, 'B': 0,
        }
        self.options['threshold'] = 50
        self.options['transparency_color'] = None

    @property
    def threshold(self):
        """get threshold for B&W conversion"""
        return self.options['threshold']

    @threshold.setter
    def threshold(self, threshold):
        """set B&W threshold for conversion """
        self.options['threshold'] = threshold

    @property
    def transparency_color(self):
        """get transparency color"""
        return self.options['transparency_color']

    @transparency_color.setter
    def transparency_color(self, transparency_color):
        """set transparency color """
        self.options['transparency_color'] = transparency_color

    def draw_pixel(self, pos_x, pos_y):
        """dummy fuction"""
        pass

    def draw_line(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """dummy fuction"""
        pass

    def draw_rect(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw a rectangle"""
        self.draw_line(pos_x1, pos_y1, pos_x2, pos_y1)
        self.draw_line(pos_x1, pos_y2, pos_x2, pos_y2)
        self.draw_line(pos_x1, pos_y1, pos_x1, pos_y2)
        self.draw_line(pos_x2, pos_y1, pos_x2, pos_y2)

    def draw_circle(self, pos_x, pos_y, radius):
        """draw a circle"""
        err = 0
        offset_x = radius
        offset_y = 0
        while offset_x >= offset_y:
            self.draw_pixel(pos_x + offset_x, pos_y + offset_y)
            self.draw_pixel(pos_x + offset_y, pos_y + offset_x)
            self.draw_pixel(pos_x - offset_y, pos_y + offset_x)
            self.draw_pixel(pos_x - offset_x, pos_y + offset_y)
            self.draw_pixel(pos_x - offset_x, pos_y - offset_y)
            self.draw_pixel(pos_x - offset_y, pos_y - offset_x)
            self.draw_pixel(pos_x + offset_y, pos_y - offset_x)
            self.draw_pixel(pos_x + offset_x, pos_y - offset_y)
            if err <= 0:
                offset_y += 1
                err += 2*offset_y + 1
            else:
                offset_x -= 1
                err -= 2*offset_x + 1

    def draw_arc(self, pos_x, pos_y, radius, start, end):
        """draw an arc"""
        start = start * math.pi / 180
        end = end * math.pi / 180

        err = 0
        offset_x = radius
        offset_y = 0
        while offset_x >= offset_y:
            if start <= math.atan2(offset_y, offset_x) <= end:
                self.draw_pixel(pos_x + offset_x, pos_y + offset_y)
            if start <= math.atan2(offset_x, offset_y) <= end:
                self.draw_pixel(pos_x + offset_y, pos_y + offset_x)
            if start <= math.atan2(offset_x, -offset_y) <= end:
                self.draw_pixel(pos_x - offset_y, pos_y + offset_x)
            if start <= math.atan2(offset_y, -offset_x) <= end:
                self.draw_pixel(pos_x - offset_x, pos_y + offset_y)

            if start <= math.atan2(-offset_y, -offset_x) + 2*math.pi <= end:
                self.draw_pixel(pos_x - offset_x, pos_y - offset_y)
            if start <= math.atan2(-offset_x, -offset_y) + 2*math.pi <= end:
                self.draw_pixel(pos_x - offset_y, pos_y - offset_x)
            if start <= math.atan2(-offset_x, offset_y) + 2*math.pi <= end:
                self.draw_pixel(pos_x + offset_y, pos_y - offset_x)
            if start <= math.atan2(-offset_y, offset_x) + 2*math.pi <= end:
                self.draw_pixel(pos_x + offset_x, pos_y - offset_y)

            if err <= 0:
                offset_y += 1
                err += 2*offset_y + 1
            else:
                offset_x -= 1
                err -= 2*offset_x + 1



"""SPI+GPIO connection driver for SSD1306"""
import time
import spidev  # pylint: disable=I0011,F0401
import RPi.GPIO as GPIO  # pylint: disable=I0011,F0401
from gfxlcd.abstract.driver import Driver
GPIO.setmode(GPIO.BCM)


class SPI(Driver):
    """SPI driver"""
    def __init__(self):
        self.pins = {
            'RST': 13,
            'DC': 6,
        }
        self.spi = None

    def init(self):
        """init sequence"""
        for pin in self.pins:
            GPIO.setup(self.pins[pin], GPIO.OUT)
            GPIO.output(self.pins[pin], 0)

        spi = spidev.SpiDev()
        spi.open(0, 0)
        spi.max_speed_hz = 8000000
        spi.mode = 0
        self.spi = spi

    def reset(self):
        """reset device"""
        GPIO.output(self.pins['RST'], 1)
        time.sleep(0.025)
        GPIO.output(self.pins['RST'], 0)
        time.sleep(0.025)
        GPIO.output(self.pins['RST'], 1)
        time.sleep(0.025)

    def cmd(self, data, enable=None):
        """send command to device"""
        GPIO.output(self.pins['DC'], 0)
        self.spi.xfer2([data])

    def data(self, data, enable=None):
        """send data to device"""
        GPIO.output(self.pins['DC'], 1)
        self.spi.xfer2([data])
        GPIO.output(self.pins['DC'], 0)



"""SSD1306 chip driver"""
from gfxlcd.drawing.page import Page
from gfxlcd.abstract.chip import Chip


class SSD1306(Page, Chip):
    """Class for an LCD with SSD306 chip"""
    def __init__(self, width, height, driver, auto_flush=True):
        Chip.__init__(self, width, height, driver, auto_flush)
        Page.__init__(self, driver)

    def init(self):
        """inits a device"""
        self.driver.init()
        Page.init(self)
        Chip.init(self)
        self.driver.reset()
        self.driver.cmd(0xae)  # turn off panel
        self.driver.cmd(0x00)  # set low column address
        self.driver.cmd(0x10)  # set high column address
        self.driver.cmd(0x40)  # set start line address

        self.driver.cmd(0x20)  # addr mode
        self.driver.cmd(0x02)  # horizontal

        self.driver.cmd(0xb0)  # set page address
        self.driver.cmd(0x81)  # set contrast control register
        self.driver.cmd(0xff)
        self.driver.cmd(0xa1)  # a0/a1, a1 = segment 127 to 0, a0:0 to seg127
        self.driver.cmd(0xc8)  # c8/c0 set com(N-1)to com0  c0:com0 to com(N-1)
        self.driver.cmd(0xa6)  # set normal display, a6 - normal, a7 - inverted

        self.driver.cmd(0xa8)  # set multiplex ratio(16to63)
        self.driver.cmd(0x3f)  # 1/64 duty

        self.driver.cmd(0xd3)  # set display offset
        self.driver.cmd(0x00)  # not offset

        # set display clock divide ratio/oscillator frequency
        self.driver.cmd(0xd5)
        self.driver.cmd(0x80)  # set divide ratio

        self.driver.cmd(0xd9)  # set pre-charge period
        self.driver.cmd(0xf1)
        self.driver.cmd(0xda)  # set com pins hardware configuration
        self.driver.cmd(0x12)

        self.driver.cmd(0xdb)  # set vcomh
        self.driver.cmd(0x40)

        self.driver.cmd(0x8d)  # charge pump
        self.driver.cmd(0x14)  # enable charge pump
        self.driver.cmd(0xaf)  # turn on panel

    def _converted_background_color(self):
        """convert RGB background to available color"""
        return 1

    def _converted_color(self):
        """convert RGB color to available color"""
        return 1

    def flush(self, force=None):
        """flush buffer to device
        :force - boolean|None"""
        if force is None:
            force = self.options['auto_flush']

        if force:
            for j in range(0, self.height//8):
                self.set_area(0, j, self.width-1, j+1)
                for i in range(0, self.width):
                    self.driver.data(self.get_page_value(i, j))

    def set_area(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """set area to work on"""
        self.driver.cmd(0x22)
        self.driver.cmd(0xb0 + pos_y1)
        self.driver.cmd(0xb0 + pos_y2)
        self.driver.cmd(0x21)
        self.driver.cmd(pos_x1)
        self.driver.cmd(pos_x2)



"""Chip interface"""
import abc


class Chip(metaclass=abc.ABCMeta):
    """Chip class"""
    def __init__(self, width, height, driver, auto_flush):
        self.options = {}
        self._width = width
        self._height = height
        self.driver = driver
        self.options['auto_flush'] = auto_flush

    @property
    def width(self):
        """get width"""
        return self._width

    @property
    def height(self):
        """get height"""
        return self._height

    @abc.abstractmethod
    def _converted_background_color(self):
        """convert RGB background to available color"""
        pass

    @abc.abstractmethod
    def _converted_color(self):
        """convert RGB color to available color"""
        pass

    @property
    def color(self):
        """get RGB colour"""
        return self.options['color']

    @color.setter
    def color(self, rgb):
        """set (R, G, B) colour """
        self.options['color'] = {
            'R': rgb[0], 'G': rgb[1], 'B': rgb[2]
        }

    @property
    def background_color(self):
        """get background colour"""
        return self.options['background_color']

    @background_color.setter
    def background_color(self, rgb):
        """set (R, G, B) background colour """
        self.options['background_color'] = {
            'R': rgb[0], 'G': rgb[1], 'B': rgb[2]
        }

    @property
    def auto_flush(self):
        """get auto_flush"""
        return self.options['auto_flush']

    @auto_flush.setter
    def auto_flush(self, value):
        """set auto_flush"""
        self.options['auto_flush'] = bool(value)

    @abc.abstractmethod
    def init(self):
        """init a chipset"""
        pass

    @abc.abstractmethod
    def draw_pixel(self, pos_x, pos_y):
        """draw a pixel at x,y"""
        pass

    @abc.abstractmethod
    def draw_line(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw a line from point x1,y1 to x2,y2"""
        pass

    @abc.abstractmethod
    def draw_rect(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw a rectangle"""
        pass

    @abc.abstractmethod
    def draw_circle(self, pos_x, pos_y, radius):
        """draw a circle"""
        pass

    @abc.abstractmethod
    def draw_arc(self, pos_x, pos_y, radius, start, end):
        """draw an arc"""
        pass

    @abc.abstractmethod
    def fill_rect(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw a filled rectangle"""
        pass

    @abc.abstractmethod
    def draw_image(self, pos_x, pos_y, image):
        """draw a PIL image"""
        pass



"""Interface for communication driver"""
import abc


class Driver(metaclass=abc.ABCMeta):
    """Driver interface"""
    @abc.abstractmethod
    def init(self):
        """initialize a device"""
        pass

    @abc.abstractmethod
    def reset(self):
        """resets a device"""
        pass

    @abc.abstractmethod
    def cmd(self, data, enable):
        """sends command to device"""
        pass

    @abc.abstractmethod
    def data(self, data, enable):
        """sends data to device"""
        pass



import abc
from gfxlcd.drawing.pixel import Pixel


class Page(Pixel, metaclass=abc.ABCMeta):
    """Page drawing algorithm"""
    def __init__(self, driver):
        self.driver = driver
        Pixel.__init__(self, driver)
        self.buffer = []

    def init(self):
        """init page"""
        self.buffer = [[0] * (self.height // 8) for x in range(self.width)]

    def draw_pixel(self, pos_x, pos_y):
        """draw a pixel at x,y"""
        self.buffer[pos_x][pos_y//8] |= 1 << (pos_y % 8)
        self.flush()

    def _calculate_steps(self, length, step, required_length):
        """calculate lineparts - helper"""
        steps = [length for _ in range(0, step)]
        if step * length < required_length:
            for idx in range(0, required_length - step * length):
                steps[idx] += 1

        return steps

    def draw_line(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw diagonal line"""
        width = abs(pos_x2 - pos_x1)
        height = abs(pos_y2 - pos_y1)
        if pos_x1 == pos_x2:
            steps = [height]
            horizontal = False
            offset_x = offset_y = 0
        elif pos_y1 == pos_y2:
            steps = [width]
            horizontal = True
            offset_x = offset_y = 0
        elif width > height:
            if pos_x2 < pos_x1:
                pos_x1, pos_x2 = pos_x2, pos_x1
                pos_y1, pos_y2 = pos_y2, pos_y1
            offset_y = 1 if pos_y2 > pos_y1 else -1
            offset_x = 1 if pos_x2 > pos_x1 else -1
            horizontal = True
            step = height
            length = width / step
            steps = self._calculate_steps(length, step, width)

        else:
            if pos_y2 < pos_y1:
                pos_x1, pos_x2 = pos_x2, pos_x1
                pos_y1, pos_y2 = pos_y2, pos_y1
            offset_y = 1 if pos_y2 > pos_y1 else -1
            offset_x = 1 if pos_x2 > pos_x1 else -1
            horizontal = False
            step = width
            length = height / step
            steps = self._calculate_steps(length, step, height)
        delta_y = 0
        delta_x = 0
        for idx, step in enumerate(steps):
            if horizontal:
                for appendix in range(int(step)+1):
                    self.draw_pixel(
                        int(pos_x1 + delta_x + appendix),
                        int(pos_y1 + (idx * offset_y))
                    )
                delta_x += step * offset_x
            else:
                for appendix in range(int(step)+1):
                    self.draw_pixel(
                        int(pos_x1 + (idx * offset_x)),
                        int(pos_y1 + delta_y + appendix)
                    )
                delta_y += step * offset_y

    def fill_rect(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw a filled rectangle"""
        if pos_y2 < pos_y1:
            pos_y1, pos_y2 = pos_y2, pos_y1
        if pos_x2 < pos_x1:
            pos_x1, pos_x2 = pos_x2, pos_x1
        start_page = pos_y1 // 8
        start_bit = pos_y1 % 8
        end_page = pos_y2 // 8
        end_bit = pos_y2 % 8
        rows = []
        first_page = int(('0' * start_bit).rjust(8, '1'), 2)
        last_page = int('1' * (end_bit+1), 2)
        if start_page != end_page:
            rows.append(first_page)
            for _ in range(end_page - start_page - 1):
                rows.append(255)
            rows.append(last_page)
        else:
            rows.append(first_page & last_page)

        page = start_page
        for value in rows:
            for x_diff in range(pos_x2-pos_x1+1):
                self.buffer[pos_x1+x_diff][page] |= value
            page += 1

    def get_page_value(self, column, page):
        """returns value"""
        return self.buffer[column][page]

    @abc.abstractmethod
    def flush(self, force=None):
        """flush buffer to the screen"""
        pass

    def draw_image(self, pos_x, pos_y, image):
        """draw a PIL image"""
        image_file = image.convert('L')
        width, height = image_file.size
        offset_x = 0
        offset_y = 0
        for stream in list(image_file.getdata()):
            if stream > self.options['threshold'] \
                    and not self._is_transparent(stream):
                self.draw_pixel(pos_x + offset_x, pos_y + offset_y)
            offset_x += 1
            if offset_x > width - 1:
                offset_x = 0
                offset_y += 1

    def _is_transparent(self, color):
        """check if color is a transparency color"""
        if type(self.options['transparency_color']) == int \
                and color == self.options['transparency_color']:
            return True
        elif type(self.options['transparency_color']) == list \
                and color in self.options['transparency_color']:
            return True

        return False



"""Area drawing algorithm"""
import itertools
from gfxlcd.drawing.pixel import Pixel


class Area(Pixel):
    """Area drawing algorithm"""
    def __init__(self, driver):
        self.driver = driver
        Pixel.__init__(self, driver)

    def init(self):
        """additional initialization"""
        pass

    def draw_pixel(self, pos_x, pos_y):
        """draw one pixel"""
        self._set_area(pos_x, pos_y, pos_x, pos_y)
        self.driver.data(self._converted_color(), None)

    def _set_area(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """select area to work with"""
        self.driver.cmd(0x0020, None)
        self.driver.data(pos_x1, None)
        self.driver.cmd(0x0021, None)
        self.driver.data(pos_y1, None)
        self.driver.cmd(0x0050, None)
        self.driver.data(pos_x1, None)
        self.driver.cmd(0x0052, None)
        self.driver.data(pos_y1, None)
        self.driver.cmd(0x0051, None)
        self.driver.data(pos_x2, None)
        self.driver.cmd(0x0053, None)
        self.driver.data(pos_y2, None)
        self.driver.cmd(0x0022, None)

    def _draw_vertical_line(self, pos_x, pos_y, length):
        """draw vertical line"""
        self._set_area(pos_x, pos_y, pos_x, pos_y + length)
        color = self._converted_color()
        for _ in itertools.repeat(None, length):
            self.driver.data(color, None)

    def _draw_horizontal_line(self, pos_x, pos_y, length):
        """draw horizontal line"""
        self._set_area(pos_x, pos_y, pos_x + length, pos_y)
        color = self._converted_color()
        for _ in itertools.repeat(None, length):
            self.driver.data(color, None)

    def _calculate_steps(self, length, step, required_length):
        """calculate lineparts - helper"""
        steps = [length for _ in range(0, step)]
        if step * length < required_length:
            for idx in range(0, required_length - step * length):
                steps[idx] += 1

        return steps

    def draw_line(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """draw diagonal line"""
        width = abs(pos_x2 - pos_x1)
        height = abs(pos_y2 - pos_y1)
        if pos_x1 == pos_x2:
            steps = [height]
            horizontal = False
            offset_x = offset_y = 0
        elif pos_y1 == pos_y2:
            steps = [width]
            horizontal = True
            offset_x = offset_y = 0
        elif width > height:
            if pos_x2 < pos_x1:
                pos_x1, pos_x2 = pos_x2, pos_x1
                pos_y1, pos_y2 = pos_y2, pos_y1
            offset_y = 1 if pos_y2 > pos_y1 else -1
            offset_x = 1 if pos_x2 > pos_x1 else -1
            horizontal = True
            step = height
            length = width / step
            steps = self._calculate_steps(length, step, width)

        else:
            if pos_y2 < pos_y1:
                pos_x1, pos_x2 = pos_x2, pos_x1
                pos_y1, pos_y2 = pos_y2, pos_y1
            offset_y = 1 if pos_y2 > pos_y1 else -1
            offset_x = 1 if pos_x2 > pos_x1 else -1
            horizontal = False
            step = width
            length = height / step
            steps = self._calculate_steps(length, step, height)

        delta_y = 0
        delta_x = 0
        for idx, step in enumerate(steps):
            if horizontal:
                self._draw_horizontal_line(
                    int(pos_x1 + delta_x),
                    int(pos_y1 + (idx * offset_y)),
                    int(step)
                )
                delta_x += step * offset_x
            else:
                self._draw_vertical_line(
                    int(pos_x1 + (idx * offset_x)),
                    int(pos_y1 + delta_y),
                    int(step)
                )
                delta_y += step * offset_y

    def fill_rect(self, pos_x1, pos_y1, pos_x2, pos_y2):
        """fill an area"""
        size = (abs(pos_x2 - pos_x1) + 1) * (abs(pos_y2 - pos_y1) + 1)
        self._set_area(
            min(pos_x1, pos_x2),
            min(pos_y1, pos_y2),
            max(pos_x1, pos_x2),
            max(pos_y1, pos_y2)
        )
        color = self._converted_background_color()
        for _ in range(0, size):
            self.driver.data(color, None)

    def draw_image(self, pos_x, pos_y, image):
        """draw a PIL image"""
        image_file = image.convert('RGB')
        width, height = image_file.size
        self._set_area(
            pos_x,
            pos_y,
            pos_x + width - 1,
            pos_y + height - 1
        )
        row = 0
        col = 0
        area = None
        temporary_area = None
        for red, green, blue in list(image_file.getdata()):
            if self._is_transparent((red, green, blue)):
                area = (
                    pos_x,
                    pos_y + row + 1,
                    pos_x + width - 1,
                    pos_y + height - 1
                )
                temporary_area = (
                    pos_x + col + 1,
                    pos_y + row,
                    pos_x + width - 1,
                    pos_y + row
                )
            else:
                if temporary_area is not None:
                    self._set_area(*temporary_area)
                    temporary_area = None
                self.color = (red, green, blue)
                self.driver.data(self._converted_color(), None)

            col += 1
            if col > width - 1:
                col = 0
                row += 1
                if area is not None:
                    self._set_area(*area)
                    area = None

    def _is_transparent(self, color):
        """check if color is a transparency color"""
        if self.options['transparency_color'] is None:
            return False
        elif type(self.options['transparency_color'][0]) == int \
            and color == self.options['transparency_color']:
                return True
        elif type(self.options['transparency_color'][0]) == list \
            and color in self.options['transparency_color']:
                return True

        return False



"""GPIO communication driver"""
import time
import RPi.GPIO  # pylint: disable=I0011,F0401
from gfxlcd.abstract.driver import Driver
RPi.GPIO.setmode(RPi.GPIO.BCM)


class GPIO(Driver):
    """GPIO communication driver"""
    def __init__(self):
        self.pins = {
            'RS': 27,
            'W': 17,
            'DB8': 22,
            'DB9': 23,
            'DB10': 24,
            'DB11': 5,
            'DB12': 12,
            'DB13': 16,
            'DB14': 20,
            'DB15': 21,
            'RST': 25,
        }
        self.data_pins = [
            'DB8', 'DB9', 'DB10', 'DB11', 'DB12', 'DB13', 'DB14', 'DB15',
        ]

    def init(self):
        """initialize pins"""
        for pin in self.pins:
            RPi.GPIO.setup(self.pins[pin], RPi.GPIO.OUT)
            RPi.GPIO.output(self.pins[pin], 0)

    def reset(self):
        """reset a display"""
        RPi.GPIO.output(self.pins['RST'], 1)
        time.sleep(0.005)
        RPi.GPIO.output(self.pins['RST'], 0)
        time.sleep(0.005)
        RPi.GPIO.output(self.pins['RST'], 1)
        time.sleep(0.005)

    def _set_pins(self, bits):
        """set rpi pins"""
        for pin in self.data_pins:
            value = bits & 0x01
            RPi.GPIO.output(self.pins[pin], value)
            bits >>= 1

    def cmd(self, data, enable):
        """send command to display"""
        RPi.GPIO.output(self.pins['RS'], 0)
        self.send(data, enable)

    def send(self, char, enable):
        """send 16bit as 2*8bit"""
        self._set_pins(char >> 8)
        RPi.GPIO.output(self.pins['W'], 0)
        RPi.GPIO.output(self.pins['W'], 1)
        self._set_pins(char)
        RPi.GPIO.output(self.pins['W'], 0)
        RPi.GPIO.output(self.pins['W'], 1)

    def data(self, data, enable):
        """send data to display"""
        RPi.GPIO.output(self.pins['RS'], 1)
        self.send(data, enable)
